﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Globalization;
using System.Reflection;
using System.Text;
using Produmex.Foundation;
using Produmex.Foundation.Barcode;
using Produmex.Foundation.Convertors;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;

namespace WorkflowScript_EnterPickingQuantityHookFlow
{
    /// <summary>
    /// 
    /// </summary>
    public class WorkflowScript_EnterPickingQuantityHookFlow : WorkflowInstanceScriptBase
    {
        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        private struct Defs
        {
            public const int PacksInBox = 100;

            public const string ActivePickListDocEntry = "ActivePickListDocEntry";
        }

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

        public ReadOnlyBinder<PickingItemInfo> PickingItemInfo;
        public ReadOnlyBinder<double> MaximumQuantity;

        // Output parameters * do not change *
        public ReadWriteBinder<IPreAndPostActionInTransaction> PreAndPostActionObject;
        public ReadWriteBinder<double> QuantityEntered;
        public ReadWriteBinder<bool> BackRequested;

        // Sub-flows
        // <none>

        // Table definition for PMX_TEOX_SERNUM
        private class PmxTeoxaneSerialNumbersTable
        {
            public const string NAME = "PMX_TEOX_SERNUM";

            public class Columns
            {
                public static class BatchNumber
                {
                    public const string NAME = "LotNr";
                }

                public static class FirstSeqNr
                {
                    public const string NAME = "FirstSeqNr";
                }

                public static class LastSeqNr
                {
                    public const string NAME = "LastSeqNr";
                }

                public static class Quantity
                {
                    public const string NAME = "Quantity";
                }

                public static class PickListDocEntry
                {
                    public const string NAME = "PickListDocEntry";
                }

                public static class PickListLineNum
                {
                    public const string NAME = "PickListLineNum";
                }

            }
        }

        // Action object holding data and code for processing within transaction used for PickItem(...)
        private class ActionObject : IPreAndPostActionInTransaction
        {
            private class RangeQty
            {
                public int FirstSeqNr;
                public int LastSeqNr;
                public int Quantity;

                public RangeQty( int firstSeqNr, int lastSeqNr, int quantity )
                {
                    this.FirstSeqNr = firstSeqNr;
                    this.LastSeqNr = lastSeqNr;
                    this.Quantity = quantity;
                }
            }
            private List<RangeQty> m_rangeQtyBoxes = new List<RangeQty>();  // List of scanned boxes
            private List<RangeQty> m_rangeQtyPacks = new List<RangeQty>();  // List of scanned packs

            public PickingItemInfo PickingItemInfo;

            /// <summary>
            /// Check if given range is free (not yet in list)
            /// </summary>
            /// <param name="firstSeqNr">The first seq nr.</param>
            /// <param name="lastSeqNr">The last seq nr.</param>
            /// <returns></returns>
            public bool RangeIsFree( int firstSeqNr, int lastSeqNr )
            {
                foreach ( RangeQty rq in m_rangeQtyBoxes )
                {
                    if ( rq.FirstSeqNr > lastSeqNr ) break;  // Current range past test range
                    if ( rq.LastSeqNr >= firstSeqNr ) return false;  // Current range overlaps test range, not free!
                }
                foreach ( RangeQty rq in m_rangeQtyPacks )
                {
                    if ( rq.FirstSeqNr > lastSeqNr ) break;  // Current range past test range
                    if ( rq.LastSeqNr >= firstSeqNr ) return false;  // Current range overlaps test range, not free!
                }
                return true;
            }

            /// <summary>
            /// Adds the range. Boxes are added separatly, pack serial numbers are joined with adjoining existing ranges
            /// </summary>
            /// <param name="firstSeqNr">The first seq nr.</param>
            /// <param name="lastSeqNr">The last seq nr.</param>
            /// <param name="quantity">The quantity.</param>
            public void AddRange( int firstSeqNr, int lastSeqNr, int quantity )
            {
                if ( firstSeqNr != lastSeqNr )
                {
                    // Add box in numerical order of serial number
                    for ( int i = 0; i < m_rangeQtyBoxes.Count; i++ )
                    {
                        // Look for the first existing range (box) with a first-seq-nr past our last-seq-nr
                        if ( m_rangeQtyBoxes[ i ].FirstSeqNr > lastSeqNr )
                        {
                            // Past the new range, insert new range (box) before this range (box)
                            m_rangeQtyBoxes.Insert( i, new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
                            return;
                        }
                    }

                    // Add as first range in list or at the back of the list
                    m_rangeQtyBoxes.Add( new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
                    return;
                }

                // Add single pack in numerical order of serial number
                // Join adjoinging ranges of serial numbers
                for ( int i = 0; i < m_rangeQtyPacks.Count; i++ )
                {
                    if ( m_rangeQtyPacks[ i ].LastSeqNr + 1 == firstSeqNr )  // Range ends where our new range begins
                    {
                        // Join new range at end of pointed range
                        m_rangeQtyPacks[ i ].LastSeqNr = lastSeqNr;
                        m_rangeQtyPacks[ i ].Quantity += quantity;

                        if ( i + 1 < m_rangeQtyPacks.Count )
                        {
                            // Join new range at begin of next range as well
                            if ( m_rangeQtyPacks[ i + 1 ].FirstSeqNr == lastSeqNr + 1 )
                            {
                                m_rangeQtyPacks[ i ].LastSeqNr = m_rangeQtyPacks[ i + 1 ].LastSeqNr;
                                m_rangeQtyPacks[ i ].Quantity += m_rangeQtyPacks[ i + 1 ].Quantity;
                                m_rangeQtyPacks.RemoveAt( i + 1 );
                            }
                        }

                        return;
                    }

                    if ( m_rangeQtyPacks[ i ].FirstSeqNr == lastSeqNr + 1 )  // Range begins where our new range ends
                    {
                        // Join new range at begin of pointed range
                        m_rangeQtyPacks[ i ].FirstSeqNr = firstSeqNr;
                        m_rangeQtyPacks[ i ].Quantity += quantity;
                        return;
                    }

                    if ( m_rangeQtyPacks[ i ].FirstSeqNr > lastSeqNr + 1 )
                    {
                        // Past the new range, so add as new range
                        m_rangeQtyPacks.Insert( i, new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
                        return;
                    }
                }

                // Add as first or at the back of the list
                m_rangeQtyPacks.Add( new RangeQty( firstSeqNr, lastSeqNr, quantity ) );
            }

            /// <summary>
            /// Reduces a range in the database -> remove/extract a range from an existing range
            /// If the range to remove:
            ///  * falls within the existing range, do nothing, we cannot split a range in the database
            ///  * overlaps with the front of the existing range, cut from the front of the existing range
            ///  * overlaps with the back of the existing range, cut from the back of the existing range
            /// Cutting off of front or back also includes earlier ranges that would have required a 'split' (first case in list above)
            /// </summary>
            /// <param name="firstSeqNr">The first seq nr.</param>
            /// <param name="lastSeqNr">The last seq nr.</param>
            /// <param name="rowQty">The row qty.</param>
            /// <param name="removeFirstSeqNr">The remove first seq nr.</param>
            /// <param name="removeLastSeqNr">The remove last seq nr.</param>
            /// <param name="dbConn">The db conn.</param>
            /// <returns></returns>
            private int ReduceRange( int firstSeqNr, int lastSeqNr, int rowQty, int removeFirstSeqNr, int removeLastSeqNr, Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                if ( removeFirstSeqNr > firstSeqNr && removeLastSeqNr < lastSeqNr )
                {
                    // Removing inner part, do nothing, we cannot split (which quantity for which part in case of a range with missing serial numbers)
                    return removeLastSeqNr - removeFirstSeqNr + 1;
                }

                if ( removeLastSeqNr < lastSeqNr )  // Removing front part
                {
                    int newFirstSeqNr = removeLastSeqNr + 1;
                    int overlappedQty = newFirstSeqNr - firstSeqNr;
                    int reduceQty = overlappedQty;

                    // Scan remaining range for already existing range assignments and remove them as well 
                    string query = BuildQuery.GetSerialNumberData( this.PickingItemInfo.BatchNumber, removeLastSeqNr, lastSeqNr, true );
                    using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                    {
                        while ( !rs.EoF )
                        {
                            int currentFirstSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ] );
                            if ( currentFirstSeqNr > newFirstSeqNr ) break;  // Beyond join/overlap

                            int currentLastSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ] );
                            if ( currentLastSeqNr >= newFirstSeqNr
                                && ( currentFirstSeqNr != firstSeqNr || currentLastSeqNr != lastSeqNr ) )
                            {
                                // Skip over adjoining range
                                newFirstSeqNr = Math.Min( currentLastSeqNr + 1, lastSeqNr + 1 );
                                reduceQty += newFirstSeqNr - currentFirstSeqNr;
                            }

                            rs.MoveNext();
                        }
                    }

                    if ( newFirstSeqNr <= lastSeqNr )
                    {
                        query = BuildQuery.MoveResizeRange( this.PickingItemInfo.BatchNumber, firstSeqNr, newFirstSeqNr, lastSeqNr, rowQty - reduceQty );
                    }
                    else
                    {
                        // Nothing left of db range, remove it completely
                        query = BuildQuery.DeleteRange( this.PickingItemInfo.BatchNumber, firstSeqNr );
                    }
                    SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                    return overlappedQty;
                }
                else  // Removing back part
                {
                    int newLastSeqNr = removeFirstSeqNr - 1;
                    int overlappedQty = lastSeqNr - newLastSeqNr;
                    int reduceQty = overlappedQty;

                    // Scan remaining range for already existing range assignments and remove them as well
                    string query = BuildQuery.GetSerialNumberData( this.PickingItemInfo.BatchNumber, firstSeqNr, removeFirstSeqNr, true );
                    using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                    {
                        rs.MoveLast();
                        while ( !rs.BoF )
                        {
                            int currentLastSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ] );
                            if ( currentLastSeqNr < newLastSeqNr ) break;  // Beyond join/overlap

                            int currentFirstSeqNr = Convert.ToInt32( rs[ PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ] );
                            if ( currentFirstSeqNr <= newLastSeqNr 
                                && ( currentFirstSeqNr != firstSeqNr || currentLastSeqNr != lastSeqNr ) )
                            {
                                // Skip over adjoining range
                                newLastSeqNr = Math.Max( firstSeqNr - 1, currentFirstSeqNr - 1 );
                                reduceQty += currentLastSeqNr - newLastSeqNr;
                            }

                            rs.MovePrevious();
                        }
                    }

                    if ( firstSeqNr <= newLastSeqNr )
                    {
                        query = BuildQuery.MoveResizeRange( this.PickingItemInfo.BatchNumber, firstSeqNr, firstSeqNr, newLastSeqNr, rowQty - reduceQty );
                    }
                    else
                    {
                        // Nothing left of db range, remove it completely
                        query = BuildQuery.DeleteRange( this.PickingItemInfo.BatchNumber, firstSeqNr );
                    }
                    SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                    return overlappedQty;
                }
            }

            private void UpdateRange( RangeQty rangeQty, bool isBox, Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                string query = BuildQuery.GetSerialNumberData( this.PickingItemInfo.BatchNumber, rangeQty.FirstSeqNr, rangeQty.LastSeqNr, true );
                using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
                {
                    if ( rs.EoF )
                    {
                        // No records found
                        throw new ProdumexException( "Serial number data unexpectedly changed, items not picked!" );
                    }

                    int pendingUpdateFirstSeqNr = -1;
                    int pendingUpdateNewLastSeqNr = -1;
                    int pendingUpdateNewQuantity = -1;
                    bool pendingUpdateIsInsert = false;
                    bool dbQtyIsExact = true;
                    bool reduceAtEnd = false;
                    bool reduceAtBegin = false;

                    int dbQty = 0;
                    int lastBigRangeFirstSeqNr = -1;
                    int lastBigRangeLastSeqNr = -1;
                    while ( !rs.EoF )
                    {
                        if ( SystemInt32Convertor.ObjectToIntOrNull( rs[ Defs.ActivePickListDocEntry ] ).HasValue )
                        {
                            // Serial number already used
                            throw new ProdumexException( "Serial number data unexpectedly changed, items not picked!" );
                        }

                        int rowQty = rs.GetTypedValue<int>( PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME );
                        if ( rowQty > 0 )
                        {
                            int currentFirstSeqNr = rs.GetTypedValue<int>( PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME );
                            int currentLastSeqNr = rs.GetTypedValue<int>( PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME );

                            if ( currentFirstSeqNr >= lastBigRangeFirstSeqNr && currentLastSeqNr <= lastBigRangeLastSeqNr )
                            {
                                // Exists inside bigger range, delete it (single packs scanned and then picklist closed results in remnants)
                                query = BuildQuery.DeleteRange( this.PickingItemInfo.BatchNumber, currentFirstSeqNr );
                                SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                            }
                            else
                            {
                                lastBigRangeFirstSeqNr = currentFirstSeqNr;
                                lastBigRangeLastSeqNr = currentLastSeqNr;

                                bool reduceDone = false;

                                if ( currentFirstSeqNr < rangeQty.FirstSeqNr )  // Current range begins before wanted range
                                {
                                    if ( reduceAtEnd )
                                    {
                                        // This can only occur ones, otherwise the db range data is corrupt (overlapping ranges)
                                        throw new ProdumexException( "Serial number data in database is incorrect, items not picked!" );
                                    }
                                    reduceAtEnd = true;
                                    reduceDone = true;

                                    if ( currentLastSeqNr - currentFirstSeqNr + 1 != rowQty ) dbQtyIsExact = false;

                                    // Reduce current range
                                    int overlappedQty = ReduceRange( currentFirstSeqNr, currentLastSeqNr, rowQty,
                                        rangeQty.FirstSeqNr, rangeQty.LastSeqNr, dbConn );

                                    currentFirstSeqNr = rangeQty.FirstSeqNr;
                                    currentLastSeqNr = Math.Min( currentLastSeqNr, rangeQty.LastSeqNr );
                                    if ( pendingUpdateFirstSeqNr < 0 ) pendingUpdateIsInsert = true;

                                    rowQty = overlappedQty;
                                }
                                else if ( currentLastSeqNr > rangeQty.LastSeqNr )  // Current range ends beyond wanted range
                                {
                                    if ( reduceAtBegin )
                                    {
                                        // This can only occur ones, otherwise the db range data is corrupt (overlapping ranges)
                                        throw new ProdumexException( "Serial number data in database is incorrect, items not picked!" );
                                    }
                                    reduceAtBegin = true;
                                    reduceDone = true;

                                    if ( currentLastSeqNr - currentFirstSeqNr + 1 != rowQty ) dbQtyIsExact = false;

                                    // Reduce current range
                                    int overlappedQty = ReduceRange( currentFirstSeqNr, currentLastSeqNr, rowQty,
                                        rangeQty.FirstSeqNr, rangeQty.LastSeqNr, dbConn );

                                    currentLastSeqNr = rangeQty.LastSeqNr;
                                    if ( pendingUpdateFirstSeqNr < 0 ) pendingUpdateIsInsert = true;

                                    rowQty = overlappedQty;
                                }

                                if ( pendingUpdateFirstSeqNr >= 0 )
                                {
                                    // There is a pending update, join with it if adjoining range, otherwise execute pending update
                                    if ( pendingUpdateNewLastSeqNr + 1 != currentFirstSeqNr )
                                    {
                                        // Current range doesn't join with pending update, execute pending update
                                        if ( pendingUpdateIsInsert )
                                        {
                                            query = BuildQuery.InsertRange( this.PickingItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                                this.PickingItemInfo.PickListDocEntry, this.PickingItemInfo.PickListLineID );
                                        }
                                        else
                                        {
                                            query = BuildQuery.UpdateRange( this.PickingItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                                this.PickingItemInfo.PickListDocEntry, this.PickingItemInfo.PickListLineID );
                                        }
                                        SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );

                                        pendingUpdateFirstSeqNr = -1;
                                        pendingUpdateIsInsert = false;
                                    }
                                    else if ( !reduceDone )
                                    {
                                        // Current range does join with pending update, delete current range
                                        query = BuildQuery.DeleteRange( this.PickingItemInfo.BatchNumber, currentFirstSeqNr );
                                        SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                                    }
                                }
                                if ( pendingUpdateFirstSeqNr < 0 )
                                {
                                    pendingUpdateFirstSeqNr = currentFirstSeqNr;
                                    pendingUpdateNewQuantity = 0;
                                }
                                pendingUpdateNewLastSeqNr = currentLastSeqNr;
                                pendingUpdateNewQuantity += rowQty;

                                dbQty += rowQty;
                            }
                        }
                        else
                        {
                            // Check here for voided serial numbers
                        }

                        rs.MoveNext();
                    }
                    if ( dbQty != rangeQty.Quantity )
                    {
                        // Quantity doesn't match anymore
                        throw new ProdumexException( "Serial number data unexpectedly changed, items not picked!" );
                    }
                    if ( pendingUpdateFirstSeqNr >= 0 )
                    {
                        // There is a pending update, execute it
                        if ( pendingUpdateIsInsert )
                        {
                            query = BuildQuery.InsertRange( this.PickingItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                this.PickingItemInfo.PickListDocEntry, this.PickingItemInfo.PickListLineID );
                        }
                        else
                        {
                            query = BuildQuery.UpdateRange( this.PickingItemInfo.BatchNumber, pendingUpdateFirstSeqNr, pendingUpdateNewLastSeqNr, pendingUpdateNewQuantity,
                                this.PickingItemInfo.PickListDocEntry, this.PickingItemInfo.PickListLineID );
                        }
                        SboRecordsetHelper.RunQueryNoReturn( s_log, query, dbConn );
                    }
                }
            }

            #region IPreAndPostActionInTransaction Members

            public System.Type RequiredConnectionType
            {
                get { return null; }
            }

            public void PreAction( Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                // Now verify and update serial number ranges in database with DocEntry/LineNum of PickList

                // Update boxes
                foreach ( RangeQty rq in m_rangeQtyBoxes )
                {
                    UpdateRange( rq, true, dbConn );
                }

                // Update packs
                foreach ( RangeQty rq in m_rangeQtyPacks )
                {
                    UpdateRange( rq, false, dbConn );
                }
            }

            public void PostAction( Produmex.Foundation.Data.Sbo.PmxDbConnection dbConn )
            {
                // Do nothing
            }

            #endregion
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript_EnterPickingQuantityHookFlow"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript_EnterPickingQuantityHookFlow( WorkflowInstanceBase parent, WorkflowInstanceFactory factory )
            : base(parent, factory)
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter( "Session" ) as Session;
            ISboProviderService sboProviderService = GetScopeParameter( "<WwfService>ISboService" ) as ISboProviderService;

            string initialErrorKey = null;

            PickingItemInfo pickingItemInfo = this.PickingItemInfo.Get();
            int qtyStillToPick = (int)this.MaximumQuantity.Get();
            int qtyPicked = 0;

            ActionObject actionObject = new ActionObject();
            actionObject.PickingItemInfo = pickingItemInfo;

        Step_ScanSerialNumber:

            session.ShowScreen( typeof( Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreen ),
                this.DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "Scan box or pack to add",
                "AllowedButtons", new string[] { "MSG_BUTTON_DONE" },
                "ItemCode", pickingItemInfo.ItemCode,
                "ItemDescription", pickingItemInfo.ItemDescription,
                "BatchNumber", pickingItemInfo.BatchNumber,
                "BestBeforeDate", pickingItemInfo.BestBeforeDate,
                "LocationCode", pickingItemInfo.LocationCode,
                "StillToPick", qtyStillToPick,
                "TotalQuantityToPick", this.MaximumQuantity.Get() ) );
            Message msg = WaitForMessage();
            initialErrorKey = null;

            string lotNr = null;
            int firstSeqNr = 0;
            int lastSeqNr = 0;

            if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.DataEntered" )
            {
                BarcodeContainer data = ExtractParameter<BarcodeContainer>( msg.Parameters, "data" );
                if (string.IsNullOrEmpty(data.FreeBarcode))
                {
                    //no data was provided
                    initialErrorKey = "MSG_ERROR_INVALID_BARCODE";
                    goto Step_ScanSerialNumber;
                }
                string[] dataFields = data.FreeBarcode.Split( ';' );
                try
                {
                    if ( dataFields.Length == 5 )
                    {
                        // Single pack
                        lotNr = dataFields[ 1 ];
                        firstSeqNr = Convert.ToInt32( dataFields[ 2 ] );
                        lastSeqNr = firstSeqNr;
                        goto Step_ProcessScan;
                    }
                    else if ( dataFields.Length == 9 )
                    {
                        // Full box
                        lotNr = dataFields[ 2 ];
                        firstSeqNr = Convert.ToInt32( dataFields[ 4 ] );
                        lastSeqNr = Convert.ToInt32( dataFields[ 5 ] );
                        goto Step_ProcessScan;
                    }
                }
                catch ( Exception ex )
                {
                    initialErrorKey = ex.Message;
                    goto Step_ScanSerialNumber;
                }

                initialErrorKey = "MSG_ERROR_INVALID_BARCODE";
                goto Step_ScanSerialNumber;
            }
            if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.ButtonClicked" )
            {
                string buttonKey = ExtractParameter<string>( msg.Parameters, "buttonKey" );
                if ( buttonKey == "MSG_BUTTON_DONE" )
                {
                    this.PreAndPostActionObject.Set( actionObject );
                    this.QuantityEntered.Set( qtyPicked );
                    return;
                }
            }
            if ( msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.Picking.IIdentifyGS1DataPickingScreenReply.BackRequested" )
            {
                this.BackRequested.Set( true );
                return;
            }

            initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
            goto Step_ScanSerialNumber;

        Step_ProcessScan:

            if ( lotNr != pickingItemInfo.BatchNumber )
            {
                initialErrorKey = "MSG_ERROR_INCORRECT_BATCH_NUMBER";
                goto Step_ScanSerialNumber;
            }

            if ( !actionObject.RangeIsFree( firstSeqNr, lastSeqNr ) )
            {
                initialErrorKey = ( firstSeqNr == lastSeqNr )
                    ? "MSG_ERROR_SERIAL_NUMBER_ALREADY_SCANNED" : "MSG_ERROR_SERIAL_NUMBERS_OF_BOX_ALREADY_SCANNED";
                goto Step_ScanSerialNumber;
            }

            string query = BuildQuery.GetSerialNumberData( lotNr, firstSeqNr, lastSeqNr, false );
            DataSet dsSerialNumberData = sboProviderService.RunView( false, null, null, query );

            if ( GetNrOfRows( dsSerialNumberData ) < 1 )
            {
                initialErrorKey = "MSG_ERROR_UNKNOWN_SERIAL_NUMBERS";
                goto Step_ScanSerialNumber;
            }

            int scannedQty = 0;
            int lastBigRangeFirstSeqNr = -1;
            int lastBigRangeLastSeqNr = -1;
            foreach ( DataRow row in dsSerialNumberData.Tables[ 0 ].Rows )
            {
                if ( SystemInt32Convertor.ObjectToIntOrNull( row[ Defs.ActivePickListDocEntry ] ).HasValue )
                {
                    initialErrorKey = ( firstSeqNr == lastSeqNr )
                        ? "MSG_ERROR_SERIAL_NUMBER_ALREADY_SCANNED" : "MSG_ERROR_SERIAL_NUMBERS_OF_BOX_ALREADY_SCANNED";
                    goto Step_ScanSerialNumber;
                }
                int rowQty = SystemInt32Convertor.ObjectToInt( row[ PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ], 0, 0 );
                if ( rowQty > 0 )
                {
                    int currentFirstSeqNr = Convert.ToInt32( row[ PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ] );
                    int currentLastSeqNr = Convert.ToInt32( row[ PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ] );

                    if ( currentFirstSeqNr >= lastBigRangeFirstSeqNr && currentLastSeqNr <= lastBigRangeLastSeqNr )
                    {
                        // Exists inside bigger range, ignore it (single packs scanned and then picklist closed results in remnants)
                    }
                    else
                    {
                        lastBigRangeFirstSeqNr = currentFirstSeqNr;
                        lastBigRangeLastSeqNr = currentLastSeqNr;

                        if ( firstSeqNr == lastSeqNr )
                        {
                            scannedQty += 1;
                        }
                        else
                        {
                            if ( currentLastSeqNr - currentFirstSeqNr + 1 == rowQty  // Quantity is exact (no missing serial numbers in range)
                                || ( firstSeqNr == currentFirstSeqNr && lastSeqNr == currentLastSeqNr ) )  // Range match
                            {
                                scannedQty += Math.Min( lastSeqNr, currentLastSeqNr ) - Math.Max( firstSeqNr, currentFirstSeqNr ) + 1;
                            }
                            else
                            {
                                initialErrorKey = "MSG_ERROR_SERIAL_NUMBERS_OF_BOX_ALREADY_SCANNED";
                                goto Step_ScanSerialNumber;
                            }
                        }
                    }
                }
                else
                {
                    // For single packs we could check here for voided serial numbers (if voiding is added to the system)
                }
            }
            if ( scannedQty != ( ( firstSeqNr == lastSeqNr ) ? 1 : Defs.PacksInBox ) )
            {
                // qty seems wrong
            }
            if ( scannedQty > qtyStillToPick )
            {
                initialErrorKey = "MSG_ERROR_SCANNED_QUANTITY_TOO_HIGH";
                goto Step_ScanSerialNumber;
            }

            qtyStillToPick -= scannedQty;
            qtyPicked += scannedQty;
            actionObject.AddRange( firstSeqNr, lastSeqNr, scannedQty );
            goto Step_ScanSerialNumber;
        }

        #endregion

        private int GetNrOfRows( DataSet ds )
        {
            return ( ds != null && ds.Tables.Count > 0 ) ? ds.Tables[ 0 ].Rows.Count : 0;
        }

        static class BuildQuery
        {
            internal static string GetSerialNumberData( string lotNr, int firstSeqNr, int lastSeqNr, bool updateLock )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "SELECT {0}.*",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );
                strBuilder.Append( ", CASE" );
                {
                    strBuilder.AppendFormat( "\r\n    WHEN ( {0} <> {1}",
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LineStatus.NAME ),
                        SqlCommandHelper.FormatAndEscapeColumnValue( PmxDocumentStatusColumnValues.Closed.VALUE, CultureInfo.InvariantCulture ) );
                    strBuilder.AppendFormat( " AND {0} <> {1} )",
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListTable.NAME, PmxPickListTable.Columns.DocStatus.NAME ),
                        SqlCommandHelper.FormatAndEscapeColumnValue( PmxDocumentStatusColumnValues.Closed.VALUE, CultureInfo.InvariantCulture ) );
                    strBuilder.AppendFormat( " OR {0} IS NOT NULL",
                        SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.PmxSalesShippingDocEntry.NAME ) );
                    strBuilder.AppendFormat( " OR {0} IS NOT NULL",
                        SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.SalesDeliveryNoteLineDocEntry.NAME ) );
                    strBuilder.AppendFormat( " THEN {0}",
                        SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                }
                strBuilder.AppendFormat( "\r\n    ELSE NULL END AS {0}",
                    SqlCommandHelper.EscapeColumnName( Defs.ActivePickListDocEntry ) );

                strBuilder.AppendFormat( "\r\n FROM {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                if ( updateLock )
                {
                    strBuilder.Append( " WITH ( UPDLOCK )" );
                }

                strBuilder.AppendFormat( "\r\n LEFT JOIN ( {0}",
                    SqlCommandHelper.EscapeTableName( PmxPickListLineTable.NAME ) );
                {
                    strBuilder.AppendFormat( "\r\n    INNER JOIN {0}",
                        SqlCommandHelper.EscapeTableName( PmxPickListTable.NAME ) );
                    strBuilder.AppendFormat( " ON {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME ),
                        SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME ) );
                }
                strBuilder.AppendFormat( "\r\n ) ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.AppendFormat( " AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.LineNum.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ) );

                strBuilder.AppendFormat( "\r\n LEFT JOIN {0}",
                    SqlCommandHelper.EscapeTableName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME ) );
                strBuilder.AppendFormat( " ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.PickListLineDocEntry.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.AppendFormat( " AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxLinkSalesDeliveryNoteLinePickListLineTable.NAME, PmxLinkSalesDeliveryNoteLinePickListLineTable.Columns.PickListLineLineNum.NAME ),
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} <= {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} >= {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                strBuilder.AppendFormat( "\r\n ORDER BY {0}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ) );

                return strBuilder.ToString();
            }

            internal static string InsertRange( string lotNr, int firstSeqNr, int lastSeqNr, int quantity, int pickListDocEntry, int pickListLineNum )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "INSERT INTO {0} ( ",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ) );
                strBuilder.Append( " )" );

                strBuilder.Append( "\r\n VALUES ( " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( lastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( quantity, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, CultureInfo.InvariantCulture ) );
                strBuilder.Append( ", " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( pickListLineNum, CultureInfo.InvariantCulture ) );
                strBuilder.Append( " )" );

                return strBuilder.ToString();
            }

            internal static string UpdateRange( string lotNr, int firstSeqNr, int newLastSeqNr, int newQuantity, int pickListDocEntry, int pickListLineNum )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "UPDATE {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.AppendFormat( "\r\n SET {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newLastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newQuantity, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListDocEntry.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.PickListLineNum.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( pickListLineNum, CultureInfo.InvariantCulture ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );
                
                return strBuilder.ToString();
            }

            internal static string DeleteRange( string lotNr, int firstSeqNr )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "DELETE FROM {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }

            internal static string MoveResizeRange( string lotNr, int firstSeqNr, int newFirstSeqNr, int newLastSeqNr, int newQuantity )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat( "UPDATE {0}",
                    SqlCommandHelper.EscapeTableName( PmxTeoxaneSerialNumbersTable.NAME ) );

                strBuilder.AppendFormat( "\r\n SET {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newFirstSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.LastSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newLastSeqNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( ", {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.Quantity.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( newQuantity, CultureInfo.InvariantCulture ) );

                strBuilder.AppendFormat( "\r\n WHERE {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.BatchNumber.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( lotNr, CultureInfo.InvariantCulture ) );
                strBuilder.AppendFormat( "\r\n AND {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName( PmxTeoxaneSerialNumbersTable.NAME, PmxTeoxaneSerialNumbersTable.Columns.FirstSeqNr.NAME ),
                    SqlCommandHelper.FormatAndEscapeColumnValue( firstSeqNr, CultureInfo.InvariantCulture ) );

                return strBuilder.ToString();
            }
        }


    }
}
